@@ -87,7 +87,8 @@
 #define CMSPAR   010000000000
 #endif
 
-#define LINE_SIZE_MAX 1000
+#define MAX_LINE_LEN PATH_MAX
+#define MAX_HISTORY 1000
 
 #define KEY_0 0x30
 #define KEY_1 0x31
@@ -95,7 +96,12 @@
 #define KEY_3 0x33
 #define KEY_4 0x34
 #define KEY_5 0x35
+#define KEY_6 0x36
+#define KEY_7 0x37
+#define KEY_8 0x38
+#define KEY_9 0x39
 #define KEY_QUESTION 0x3f
+#define KEY_A 0x61
 #define KEY_B 0x62
 #define KEY_C 0x63
 #define KEY_E 0x65
@@ -110,9 +116,9 @@
 #define KEY_P 0x70
 #define KEY_Q 0x71
 #define KEY_R 0x72
+#define KEY_SHIFT_R 0x52
 #define KEY_S 0x73
 #define KEY_T 0x74
-#define KEY_U 0x55
 #define KEY_V 0x76
 #define KEY_X 0x78
 #define KEY_Y 0x79
@@ -130,6 +136,7 @@ typedef enum
     SUBCOMMAND_LINE_TOGGLE,
     SUBCOMMAND_LINE_PULSE,
     SUBCOMMAND_XMODEM,
+    SUBCOMMAND_MAP,
 } sub_command_t;
 
 const char random_array[] =
@@ -147,9 +154,6 @@ const char random_array[] =
 };
 
 bool interactive_mode = true;
-bool map_i_nl_cr = false;
-bool map_i_cr_nl = false;
-bool map_ign_cr = false;
 
 char key_hit = 0xff;
 
@@ -159,16 +163,8 @@ static struct termios tio, tio_old, stdout_new, stdout_old, stdin_new, stdin_old
 static unsigned long rx_total = 0, tx_total = 0;
 static bool connected = false;
 static bool standard_baudrate = true;
-static void (*print)(char c);
+static void (*printchar)(char c);
 static int device_fd;
-static bool map_i_ff_escc = false;
-static bool map_i_nl_crnl = false;
-static bool map_o_cr_nl = false;
-static bool map_o_nl_crnl = false;
-static bool map_o_del_bs = false;
-static bool map_o_ltu = false;
-static bool map_o_nulbrk = false;
-static bool map_o_msblsb = false;
 static char hex_chars[2];
 static unsigned char hex_char_index = 0;
 static char tty_buffer[BUFSIZ*2];
@@ -177,7 +173,7 @@ static char *tty_buffer_write_ptr = tty_buffer;
 static pthread_t thread;
 static int pipefd[2];
 static pthread_mutex_t mutex_input_ready = PTHREAD_MUTEX_INITIALIZER;
-static char line[LINE_SIZE_MAX];
+static char line[MAX_LINE_LEN];
 
 static void optional_local_echo(char c)
 {
@@ -186,7 +182,7 @@ static void optional_local_echo(char c)
         return;
     }
 
-    print(c);
+    printchar(c);
 
     if ((option.output_mode == OUTPUT_MODE_NORMAL) && (c == 127))
     {
@@ -253,7 +249,7 @@ ssize_t tty_write(int fd, const void *buffer, size_t count)
     ssize_t retval = 0, bytes_written = 0;
     size_t i;
 
-    if (map_o_ltu)
+    if (option.map_o_ltu)
     {
         // Convert lower case to upper case
         for (i = 0; i<count; i++)
@@ -585,13 +581,13 @@ static int tio_readln(void)
     char *p = line;
 
     /* Read line, accept BS and DEL as rubout characters */
-    for (p = line ; p < &line[LINE_SIZE_MAX-1]; )
+    for (p = line ; p < &line[MAX_LINE_LEN-1]; )
     {
         if (read(pipefd[0], p, 1) > 0)
         {
             if (*p == 0x08 || *p == 0x7f)
             {
-                if (p > line )
+                if (p > line)
                 {
                     write(STDOUT_FILENO, "\b \b", 3);
                     p--;
@@ -612,11 +608,11 @@ void tty_output_mode_set(output_mode_t mode)
     switch (mode)
     {
         case OUTPUT_MODE_NORMAL:
-            print = print_normal;
+            printchar = print_normal;
             break;
 
         case OUTPUT_MODE_HEX:
-            print = print_hex;
+            printchar = print_hex;
             break;
 
         case OUTPUT_MODE_END:
@@ -624,6 +620,32 @@ void tty_output_mode_set(output_mode_t mode)
     }
 }
 
+static void mappings_print(void)
+{
+    if (option.map_i_cr_nl || option.map_ign_cr || option.map_i_ff_escc ||
+        option.map_i_nl_cr || option.map_i_nl_crnl || option.map_o_cr_nl ||
+        option.map_o_del_bs || option.map_o_nl_crnl || option.map_o_ltu ||
+        option.map_o_nulbrk || option.map_o_msblsb)
+    {
+        tio_printf(" Mappings:%s%s%s%s%s%s%s%s%s%s%s",
+                option.map_i_cr_nl ? " ICRNL" : "",
+                option.map_ign_cr ? " IGNCR" : "",
+                option.map_i_ff_escc ? " IFFESCC" : "",
+                option.map_i_nl_cr ? " INLCR" : "",
+                option.map_i_nl_crnl ? " INLCRNL" : "",
+                option.map_o_cr_nl ? " OCRNL" : "",
+                option.map_o_del_bs ? " ODELBS" : "",
+                option.map_o_nl_crnl ? " ONLCRNL" : "",
+                option.map_o_ltu ? " OLTU" : "",
+                option.map_o_nulbrk ? " ONULBRK" : "",
+                option.map_o_msblsb ? " MSB2LSB" : "");
+    }
+    else
+    {
+        tio_printf(" Mappings: none");
+    }
+}
+
 void handle_command_sequence(char input_char, char *output_char, bool *forward)
 {
     char unused_char;
@@ -692,7 +714,7 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
                         {
                             tio_printf("Sending file '%s'  ", line);
                             tio_printf("Press any key to abort transfer");
-                            tio_printf("%s", xymodem_send(device_fd, line, XMODEM_CRC) < 0 ? "Aborted" : "Done");
+                            tio_printf("%s", xymodem_send(device_fd, line, XMODEM_1K) < 0 ? "Aborted" : "Done");
                         }
                         break;
 
@@ -707,11 +729,78 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
                         }
                         break;
 
+                    case KEY_2:
+                        tio_printf("Receive file with XMODEM-CRC");
+                        tio_printf_raw("Enter file name: ");
+                        if (tio_readln())
+                        {
+                            tio_printf("Ready to receiving file '%s'  ", line);
+                            tio_printf("Press any key to abort transfer");
+                            tio_printf("%s", xymodem_receive(device_fd, line, XMODEM_CRC) < 0 ? "Aborted" : "Done");
+                        }
+                        break;
+
                     default:
                         tio_error_print("Invalid protocol option");
                         break;
                 }
                 break;
+
+            case SUBCOMMAND_MAP:
+                switch (input_char)
+                {
+                    case KEY_0:
+                        option.map_i_cr_nl = !option.map_i_cr_nl;
+                        tty_reconfigure();
+                        tio_printf("ICRNL is %s", option.map_i_cr_nl ? "set" : "unset");
+                        break;
+                    case KEY_1:
+                        option.map_ign_cr = !option.map_ign_cr;
+                        tty_reconfigure();
+                        tio_printf("IGNCR is %s", option.map_ign_cr ? "set" : "unset");
+                        break;
+                    case KEY_2:
+                        option.map_i_ff_escc = !option.map_i_ff_escc;
+                        tio_printf("IFFESCC is %s", option.map_i_ff_escc ? "set" : "unset");
+                        break;
+                    case KEY_3:
+                        option.map_i_nl_cr = !option.map_i_nl_cr;
+                        tty_reconfigure();
+                        tio_printf("INLCR is %s", option.map_i_nl_cr ? "set" : "unset");
+                        break;
+                    case KEY_4:
+                        option.map_i_nl_crnl = !option.map_i_nl_crnl;
+                        tio_printf("INLCRNL is %s", option.map_i_nl_crnl ? "set" : "unset");
+                        break;
+                    case KEY_5:
+                        option.map_o_cr_nl = !option.map_o_cr_nl;
+                        tio_printf("OCRNL is %s", option.map_o_cr_nl ? "set" : "unset");
+                        break;
+                    case KEY_6:
+                        option.map_o_del_bs = !option.map_o_del_bs;
+                        tio_printf("ODELBS is %s", option.map_o_del_bs ? "set" : "unset");
+                        break;
+                    case KEY_7:
+                        option.map_o_nl_crnl = !option.map_o_nl_crnl;
+                        tio_printf("ONLCRNL is %s", option.map_o_nl_crnl ? "set" : "unset");
+                        break;
+                    case KEY_8:
+                        option.map_o_ltu = !option.map_o_ltu;
+                        tio_printf("OLTU is %s", option.map_o_ltu ? "set" : "unset");
+                        break;
+                    case KEY_9:
+                        option.map_o_nulbrk = !option.map_o_nulbrk;
+                        tio_printf("ONULBRK is %s", option.map_o_nulbrk ? "set" : "unset");
+                        break;
+                    case KEY_A:
+                        option.map_o_msblsb = !option.map_o_msblsb;
+                        tio_printf("MSB2LSB is %s", option.map_o_msblsb ? "set" : "unset");
+                        break;
+                    default:
+                        tio_error_print("Invalid input");
+                        break;
+                }
+                break;
         }
 
         sub_command = SUBCOMMAND_NONE;
@@ -749,16 +838,16 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
                 tio_printf(" ctrl-%c i       Toggle input mode", option.prefix_key);
                 tio_printf(" ctrl-%c l       Clear screen", option.prefix_key);
                 tio_printf(" ctrl-%c L       Show line states", option.prefix_key);
-                tio_printf(" ctrl-%c m       Toggle MSB to LSB bit order", option.prefix_key);
+                tio_printf(" ctrl-%c m       Change mapping of characters on input or output", option.prefix_key);
                 tio_printf(" ctrl-%c o       Toggle output mode", option.prefix_key);
                 tio_printf(" ctrl-%c p       Pulse serial port line", option.prefix_key);
                 tio_printf(" ctrl-%c q       Quit", option.prefix_key);
                 tio_printf(" ctrl-%c r       Run script", option.prefix_key);
+                tio_printf(" ctrl-%c R       Execute shell command with I/O redirected to device", option.prefix_key);
                 tio_printf(" ctrl-%c s       Show statistics", option.prefix_key);
                 tio_printf(" ctrl-%c t       Toggle line timestamp mode", option.prefix_key);
-                tio_printf(" ctrl-%c U       Toggle conversion to uppercase on output", option.prefix_key);
                 tio_printf(" ctrl-%c v       Show version", option.prefix_key);
-                tio_printf(" ctrl-%c x       Send file via Xmodem", option.prefix_key);
+                tio_printf(" ctrl-%c x       Send/Receive file via Xmodem", option.prefix_key);
                 tio_printf(" ctrl-%c y       Send file via Ymodem", option.prefix_key);
                 tio_printf(" ctrl-%c ctrl-%c  Send ctrl-%c character", option.prefix_key, option.prefix_key, option.prefix_key);
                 break;
@@ -829,12 +918,13 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
 
             case KEY_C:
                 tio_printf("Configuration:");
-                options_print();
                 config_file_print();
+                options_print();
                 if (option.rs485)
                 {
                     rs485_print_config();
                 }
+                mappings_print();
                 break;
 
             case KEY_E:
@@ -892,17 +982,33 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
                 break;
 
             case KEY_M:
-                /* Toggle bit order */
-                if (!map_o_msblsb)
-                {
-                    map_o_msblsb = true;
-                    tio_printf("Switched to reverse bit order");
-                }
-                else
-                {
-                    map_o_msblsb = false;
-                    tio_printf("Switched to normal bit order");
-                }
+                /* Change mapping of characters on input or output */
+                tio_printf("Please enter which mapping to set or unset:");
+                tio_printf(" (0) ICRNL: %s mapping CR to NL on input (unless IGNCR is set)",
+                        option.map_i_cr_nl ? "Unset" : "Set");
+                tio_printf(" (1) IGNCR: %s ignoring CR on input",
+                        option.map_ign_cr ? "Unset" : "Set");
+                tio_printf(" (2) IFFESCC: %s mapping FF to ESC-c on input",
+                        option.map_i_ff_escc ? "Unset" : "Set");
+                tio_printf(" (3) INLCR: %s mapping NL to CR on input",
+                        option.map_i_nl_cr ? "Unset" : "Set");
+                tio_printf(" (4) INLCRNL: %s mapping NL to CR-NL on input",
+                        option.map_i_nl_cr ? "Unset" : "Set");
+                tio_printf(" (5) OCRNL: %s mapping CR to NL on output",
+                        option.map_o_cr_nl ? "Unset" : "Set");
+                tio_printf(" (6) ODELBS: %s mapping DEL to BS on output",
+                        option.map_o_del_bs ? "Unset" : "Set");
+                tio_printf(" (7) ONLCRNL: %s mapping NL to CR-NL on output",
+                        option.map_o_nl_crnl ? "Unset" : "Set");
+                tio_printf(" (8) OLTU: %s mapping lowercase to uppercase on output",
+                        option.map_o_ltu ? "Unset" : "Set");
+                tio_printf(" (9) ONULBRK: %s mapping NUL to send break signal on output",
+                        option.map_o_nulbrk ? "Unset" : "Set");
+                tio_printf(" (a) MSB2LSB: %s mapping MSB bit order to LSB on output",
+                        option.map_o_msblsb ? "Unset" : "Set");
+
+                // Process next input character as sub command
+                sub_command = SUBCOMMAND_MAP;
                 break;
 
             case KEY_Q:
@@ -911,7 +1017,26 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
 
             case KEY_R:
                 /* Run script */
-                script_run(device_fd);
+                tio_printf("Run Lua script")
+                tio_printf_raw("Enter file name: ");
+                if (tio_readln())
+                {
+                    clear_line();
+                    script_run(device_fd, line);
+                }
+                else
+                {
+                    clear_line();
+                    script_run(device_fd, NULL);
+                }
+                break;
+
+            case KEY_SHIFT_R:
+                /* Execute shell command */
+                tio_printf("Execute shell command with I/O redirected to device");
+                tio_printf_raw("Enter command: ");
+                if (tio_readln())
+                    execute_shell_command(device_fd, line);
                 break;
 
             case KEY_S:
@@ -946,18 +1071,15 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
                 }
                 break;
 
-            case KEY_U:
-                map_o_ltu = !map_o_ltu;
-                break;
-
             case KEY_V:
                 tio_printf("tio v%s", VERSION);
                 break;
 
             case KEY_X:
                 tio_printf("Please enter which X modem protocol to use:");
-                tio_printf(" (0) XMODEM-1K");
-                tio_printf(" (1) XMODEM-CRC");
+                tio_printf(" (0) XMODEM-1K send");
+                tio_printf(" (1) XMODEM-CRC send");
+                tio_printf(" (2) XMODEM-CRC receive");
                 // Process next input character as sub command
                 sub_command = SUBCOMMAND_XMODEM;
                 break;
@@ -1076,7 +1198,7 @@ void stdout_configure(void)
     }
 
     /* At start use normal print function */
-    print = print_normal;
+    printchar = print_normal;
 
     /* Make sure we restore old stdout settings on exit */
     atexit(&stdout_restore);
@@ -1084,9 +1206,6 @@ void stdout_configure(void)
 
 void tty_configure(void)
 {
-    bool token_found = true;
-    char *token = NULL;
-    char *buffer;
     int status;
     speed_t baudrate;
 
@@ -1238,80 +1357,30 @@ void tty_configure(void)
     tio.c_cc[VTIME] = 0; // Inter-character timer unused
     tio.c_cc[VMIN]  = 1; // Blocking read until 1 character received
 
-    /* Configure any specified input or output mappings */
-    buffer = strdup(option.map);
-    while (token_found == true)
+    /* Configure input mappings */
+    if (option.map_i_nl_cr)
     {
-        if (token == NULL)
-        {
-            token = strtok(buffer,",");
-        }
-        else
-        {
-            token = strtok(NULL, ",");
-        }
+        tio.c_iflag |= INLCR;
+    }
+    if (option.map_ign_cr)
+    {
+        tio.c_iflag |= IGNCR;
+    }
+    if (option.map_i_cr_nl)
+    {
+        tio.c_iflag |= ICRNL;
+    }
+}
 
-        if (token != NULL)
-        {
-            if (strcmp(token,"INLCR") == 0)
-            {
-                tio.c_iflag |= INLCR;
-                map_i_nl_cr = true;
-            }
-            else if (strcmp(token,"IGNCR") == 0)
-            {
-                tio.c_iflag |= IGNCR;
-                map_ign_cr = true;
-            }
-            else if (strcmp(token,"ICRNL") == 0)
-            {
-                tio.c_iflag |= ICRNL;
-                map_i_cr_nl = true;
-            }
-            else if (strcmp(token,"OCRNL") == 0)
-            {
-                map_o_cr_nl = true;
-            }
-            else if (strcmp(token,"ODELBS") == 0)
-            {
-                map_o_del_bs = true;
-            }
-            else if (strcmp(token,"IFFESCC") == 0)
-            {
-                map_i_ff_escc = true;
-            }
-            else if (strcmp(token,"INLCRNL") == 0)
-            {
-                map_i_nl_crnl = true;
-            }
-            else if (strcmp(token, "ONLCRNL") == 0)
-            {
-                map_o_nl_crnl = true;
-            }
-            else if (strcmp(token, "OLTU") == 0)
-            {
-                map_o_ltu = true;
-            }
-            else if (strcmp(token, "ONULBRK") == 0)
-            {
-                map_o_nulbrk = true;
-            }
-            else if (strcmp(token, "MSB2LSB") == 0)
-            {
-                map_o_msblsb = true;
-            }
-            else
-            {
-                printf("Error: Unknown mapping flag %s\n", token);
-                exit(EXIT_FAILURE);
-            }
-        }
-        else
-        {
-            token_found = false;
-        }
+void tty_reconfigure(void)
+{
+    tty_configure();
+
+    if (connected)
+    {
+        /* Activate new port settings */
+        tcsetattr(device_fd, TCSANOW, &tio);
     }
-    free(buffer);
 }
 
 static bool is_serial_device(const char *format, ...)
@@ -2058,17 +2127,17 @@ void forward_to_tty(int fd, char output_char)
     int status;
 
     /* Map output character */
-    if ((output_char == 127) && (map_o_del_bs))
+    if ((output_char == 127) && (option.map_o_del_bs))
     {
         output_char = '\b';
     }
-    if ((output_char == '\r') && (map_o_cr_nl))
+    if ((output_char == '\r') && (option.map_o_cr_nl))
     {
         output_char = '\n';
     }
 
     /* Map newline character */
-    if ((output_char == '\n' || output_char == '\r') && (map_o_nl_crnl))
+    if ((output_char == '\n' || output_char == '\r') && (option.map_o_nl_crnl))
     {
         const char *crlf = "\r\n";
 
@@ -2094,8 +2163,12 @@ void forward_to_tty(int fd, char output_char)
                 else
                 {
                     /* Send output to tty device */
-                    optional_local_echo(output_char);
-                    if ((output_char == 0) && (map_o_nulbrk))
+                    if (option.input_mode != INPUT_MODE_LINE)
+                    {
+                        optional_local_echo(output_char);
+                    }
+
+                    if ((output_char == 0) && (option.map_o_nulbrk))
                     {
                         status = tcsendbreak(fd, 0);
                     }
@@ -2103,6 +2176,7 @@ void forward_to_tty(int fd, char output_char)
                     {
                         status = tty_write(fd, &output_char, 1);
                     }
+
                     if (status < 0)
                     {
                         tio_warning_printf("Could not write to tty device");
@@ -2133,19 +2207,27 @@ void forward_to_tty(int fd, char output_char)
     }
 }
 
+static void print_line(const char *string, int cursor_pos)
+{
+    clear_line();
+    print("%s", string);
+    print("\r"); // Move the cursor back to the beginning
+    for (int i = 0; i < cursor_pos; ++i)
+    {
+        print("\x1b[C"); // Move the cursor right
+    }
+}
+
 int tty_connect(void)
 {
     fd_set rdfs;           /* Read file descriptor set */
     int    maxfd;          /* Maximum file descriptor used */
     char   input_char, output_char;
     char   input_buffer[BUFSIZ] = {};
-    char   line_buffer[BUFSIZ] = {};
     static bool first = true;
     int    status;
     bool   do_timestamp = false;
     char*  now = NULL;
-    unsigned int line_index = 0;
-    static char previous_char[2] = {};
     struct timeval tval_before = {}, tval_now, tval_result;
 
     /* Open tty device */
@@ -2269,7 +2351,7 @@ int tty_connect(void)
     /* Manage script activation */
     if (option.script_run != SCRIPT_RUN_NEVER)
     {
-        script_run(device_fd);
+        script_run(device_fd, NULL);
 
         if (option.script_run == SCRIPT_RUN_ONCE)
         {
@@ -2289,6 +2371,20 @@ int tty_connect(void)
         exit(status);
     }
 
+    // Initialize readline like history
+    char *history[MAX_HISTORY];
+    int history_count = 0;
+    int history_index = 0;
+
+    for (int i = 0; i < MAX_HISTORY; ++i)
+    {
+        history[i] = NULL;
+    }
+
+    int line_length = 0;
+    int cursor_pos = 0;
+    int escape = 0;
+
     /* Input loop */
     while (true)
     {
@@ -2430,7 +2526,7 @@ int tty_connect(void)
                     }
 
                     /* Convert MSB to LSB bit order */
-                    if (map_o_msblsb)
+                    if (option.map_o_msblsb)
                     {
                         char ch = input_char;
                         input_char = 0;
@@ -2441,24 +2537,24 @@ int tty_connect(void)
                     }
 
                     /* Map input character */
-                    if ((input_char == '\n') && (map_i_nl_crnl) && (!map_o_msblsb))
+                    if ((input_char == '\n') && (option.map_i_nl_crnl) && (!option.map_o_msblsb))
                     {
-                        print('\r');
-                        print('\n');
+                        printchar('\r');
+                        printchar('\n');
                         if (option.timestamp)
                         {
                             do_timestamp = true;
                         }
                     }
-                    else if ((input_char == '\f') && (map_i_ff_escc) && (!map_o_msblsb))
+                    else if ((input_char == '\f') && (option.map_i_ff_escc) && (!option.map_o_msblsb))
                     {
-                        print('\e');
-                        print('c');
+                        printchar('\e');
+                        printchar('c');
                     }
                     else
                     {
                         /* Print received tty character to stdout */
-                        print(input_char);
+                        printchar(input_char);
                     }
 
                     /* Write to log */
@@ -2531,95 +2627,156 @@ int tty_connect(void)
                                 case INPUT_MODE_LINE:
                                     switch (input_char)
                                     {
+                                        case '\r': // Carriage return
+                                            if (line_length > 0)
+                                            {
+                                                // Save to history
+                                                if (history_count < MAX_HISTORY)
+                                                {
+                                                    history[history_count] = strndup(line, line_length);
+                                                    history_count++;
+                                                }
+                                                else
+                                            {
+                                                    free(history[0]);
+                                                    memmove(&history[0], &history[1], (MAX_HISTORY - 1) * sizeof(char*));
+                                                    history[MAX_HISTORY - 1] = strndup(line, line_length);
+                                                }
+                                            }
+
+                                            line[line_length] = '\0';
+                                            if (option.local_echo == false)
+                                            {
+                                                clear_line();
+                                            }
+                                            else
+                                            {
+                                                print("\r\n");
+                                            }
+
+                                            // Write line to tty device
+                                            tty_write(device_fd, line, line_length);
+                                            tty_sync(device_fd);
+
+                                            line_length = 0;
+                                            cursor_pos = 0;
+                                            history_index = history_count;
+                                            escape = 0;
+                                            break;
+
+                                        case 127: // Backspace
+                                            if (cursor_pos > 0)
+                                            {
+                                                memmove(&line[cursor_pos - 1], &line[cursor_pos], line_length - cursor_pos);
+                                                line_length--;
+                                                cursor_pos--;
+                                                line[line_length] = '\0';
+                                                print_line(line, cursor_pos);
+                                            }
+                                            forward = false;
+                                            escape = 0;
+                                            break;
+
                                         case 27: // Escape
+                                            escape = 1;
                                             forward = false;
                                             break;
 
                                         case '[':
-                                            if (previous_char[0] == 27)
+                                            if (escape == 1)
                                             {
-                                                forward = false;
+                                                escape = 2;
+                                            }
+                                            else
+                                            {
+                                                escape = 0;
                                             }
                                             break;
 
                                         case 'A':
-                                        case 'B':
-                                        case 'C':
-                                        case 'D':
-                                            if ((previous_char[1] == 27) && (previous_char[0] == '['))
+                                            if (escape == 2)
                                             {
-                                                // Handle arrow keys
-                                                switch (input_char)
+                                                // Up arrow
+                                                if (history_index > 0)
                                                 {
-                                                    case 'A': // Up arrow
-                                                        // Ignore
-                                                        break;
-                                                    case 'B': // Down arrow
-                                                        // Ignore
-                                                        break;
-                                                    case 'C': // Right arrow
-                                                        // Ignore
-                                                        break;
-                                                    case 'D': // Left arrow
-                                                        // Ignore
-                                                        break;
+                                                    history_index--;
+                                                    strncpy(line, history[history_index], MAX_LINE_LEN-1);
+                                                    line_length = strlen(line);
+                                                    cursor_pos = line_length;
+                                                    print_line(line, cursor_pos);
                                                 }
-                                                forward = false;
-                                            }
-                                            break;
-
-                                        case '\b':
-                                        case 127: // Backspace
-                                            if (line_index)
-                                            {
-                                                printf("\b \b"); // Destructive backspace
-                                                line_index--;
                                             }
                                             forward = false;
+                                            escape = 0;
                                             break;
 
-                                        case '\r': // Carriage return
-                                            if (option.local_echo == false)
+                                        case 'B':
+                                            if (escape == 2)
                                             {
-                                                // Delete line
-                                                int i = line_index;
-                                                while (i--)
+                                                // Down arrow
+                                                if (history_index < history_count - 1)
                                                 {
-                                                    printf("\b \b"); // Destructive backspace
+                                                    history_index++;
+                                                    strncpy(line, history[history_index], MAX_LINE_LEN-1);
+                                                    line_length = strlen(line);
+                                                    cursor_pos = line_length;
+                                                    print_line(line, cursor_pos);
+                                                }
+                                                else if (history_index == history_count - 1)
+                                                {
+                                                    history_index++;
+                                                    line_length = 0;
+                                                    cursor_pos = 0;
+                                                    line[line_length] = '\0';
+                                                    print_line(line, cursor_pos);
                                                 }
                                             }
-                                            else
-                                            {
-                                                // Preserve line, go to next line
-                                                putchar('\r');
-                                                putchar('\n');
-                                            }
+                                            forward = false;
+                                            escape = 0;
+                                            break;
 
-                                            // Write buffered line to tty device
-                                            tty_write(device_fd, line_buffer, line_index);
-                                            tty_sync(device_fd);
-                                            line_index = 0;
+                                        case 'C':
+                                            if (escape == 2)
+                                            {
+                                                // Right arrow
+                                                if (cursor_pos < line_length)
+                                                {
+                                                    cursor_pos++;
+                                                    print("\x1b[C");
+                                                }
+                                            }
+                                            forward = false;
+                                            escape = 0;
+                                            break;
+
+                                        case 'D':
+                                            if (escape == 2)
+                                            {
+                                                    // Left arrow
+                                                    if (cursor_pos > 0)
+                                                    {
+                                                        cursor_pos--;
+                                                        print("\b");
+                                                    }
+                                            }
+                                            forward = false;
+                                            escape = 0;
                                             break;
 
                                         default:
-                                            if (line_index < BUFSIZ)
+                                            if (line_length < MAX_LINE_LEN - 1)
                                             {
-                                                putchar(input_char);
-                                                print_tainted_set();
-                                                line_buffer[line_index++] = input_char;
+                                                memmove(&line[cursor_pos + 1], &line[cursor_pos], line_length - cursor_pos);
+                                                line[cursor_pos] = input_char;
+                                                line_length++;
+                                                cursor_pos++;
+                                                line[line_length] = '\0';
+                                                print_line(line, cursor_pos);
                                             }
-                                            else
-                                            {
-                                                tio_error_print("Input exceeds maximum line length. Truncating.");
-                                            }
-                                            forward = false;
+                                            escape = 0;
+                                            break;
                                     }
 
-                                    // Save 2 latest stdin input characters
-                                    previous_char[1] = previous_char[0];
-                                    previous_char[0] = input_char;
-                                    break;
-
                                 default:
                                     break;
                             }
